# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides a high level API to SQL DB servers. It sits above
the Drizzle dbapi providing boots with a simple and easy to use interface to
drizzle.

This API can be implemented for any dbapi implementation allowing boots to
speak to a variety of DB servers."""

import time
from boots.api import errors
from boots.api.nodes.node import Status
from boots.api import _, n_

_server_classes = []

def extract_field_indexes(cursor_description):
    """Returns a dictionary which maps field names to their indexes."""
    return dict(zip((field[0] for field in cursor_description), range(0, len(cursor_description))))

class Rows(list):
    """Represents a chunk of rows."""
    def __init__(self, sequence, field_indexes):
        super(Rows, self).__init__(sequence)
        self.field_indexes = field_indexes

class ResultInfo(Status):
    pass

class Server(object):
    """Represents a connection to a database server.
    
    This object is used to execute commands on a remote server."""

    def __init__(self, dbapi_module, hostname, port, auth):
        """Initialize a Server object.
        
        hostname should be a string.
        port should be a positive integer value.
        auth should be a dictionary possibly containing keys:
        'username', 'password', and 'database' mapping to their
        appropriate string values."""
        self.db = dbapi_module    
        self.hostname = hostname
        self.port = port
        self.auth = auth
        self._connection = None

    def __str__(self):
        return "{0}:{1}".format(self.hostname, self.port)

    def _connect(self):
        return self.db.connect(
            host     =  self.hostname,  
            port     =  self.port,
            username =  self.auth.get("username", ""),
            password =  self.auth.get("password", ""),
            database =  self.auth.get("database", ""))

    def connect(self):
        """Establishes database connection.

        Connect to the database server whose details were provided through
        the __init__ method."""
        try:
            self._connection = self._connect()
        except self.db.Error:
            raise errors.ConnectionError(_("Could not connect to {0}").format(self))

    @property
    def is_connected(self): 
        """Describes connection state.

        TRUE if connected, FALSE if disconnected."""
        return self._connection is not None

    def _check_connected(self):
        """Raise a ProgrammingError if the Connection has been closed."""
        if not self.is_connected:
            raise errors.ConnectionError(_("Connection closed"))

    def disconnect(self):
        self._check_connected()
        """Disconnect from the connected database server."""
        self._connection.close()
        self._connection = None
        
    def _execute(self, query):
        cursor = self._connection.cursor()
        cursor.execute(query)
        return cursor
    
    def execute(self, query):
        """Executes a query, returning an iterator of resulting rows."""
        self._check_connected()
        begin_time = time.time()
        
        try:
            cursor = self._execute(query)
        except errors.LostConnectionError as lost_e:
            # FIXME: this is NOT thread safe
            # Attempt to reconnect and retry
            yield Status(_("Lost connection to {0}. Reconnecting...").format(self))
            
            if self.is_connected:
                self.disconnect()
                            
            try:
                self.connect()
            except Exception as e:
                lost_e.reconnect_exception = e
                raise lost_e
            else:
                yield Status(_("Reconnected."))

            cursor = self._execute(query)
            
        if cursor.description:
            field_indexes = extract_field_indexes(cursor.description)
            
            while True:
                chunk = cursor.fetchmany(256)
                # FIXME: Eventually it would be good to implement chunk packeting
                # as a generic nodes feature.
                if chunk:
                    yield Rows(chunk, field_indexes)
                else:
                    break
                
        yield ResultInfo({"description": cursor.description,
                          "row_count": cursor.rowcount,
                          "begin_time": begin_time,
                          "end_time": time.time()})
        return
                
    @property
    def server_version(self):
        self._check_connected()
        return self._connection.server_version
    
    @property
    def protocol_version(self):
        self._check_connected()
        return self._connection.protocol_version

try:
    import drizzle.db
    class DrizzleServer(Server):
        def __init__(self, hostname, port, auth):
            Server.__init__(self, drizzle.db, hostname, port, auth)
        
        def _connect(self):
            connection = Server._connect(self)
            connection.autocommit = True
            return connection
        
        def _execute(self, query):
            try:
                cursor = self._connection.cursor(convert=False)
                cursor.execute(query)
            except drizzle.errors.LostConnectionError:
                raise errors.LostConnectionError(server=self)
            else:
                return cursor
            
except ImportError:
    pass
else:
    _server_classes.append(DrizzleServer)

try:
    import MySQLdb
    import MySQLdb.converters
    class MySQLServer(Server):
        def __init__(self, hostname, port, auth):
            Server.__init__(self, MySQLdb, hostname, port, auth)

        def _connect(self):
            """Establishes database connection.

            Connect to the database server whose details were provided through
            the __init__ method."""

            converter = self.db.converters.conversions.copy()
            for x in range(0,256):
                converter.pop(x, None)
            if self.auth["database"] == None:
                self.auth["database"] = ""
                
            connection = self.db.connect(
                host     =  self.hostname,  
                port     =  self.port,
                user     =  self.auth.get("username", ""),
                passwd   =  self.auth.get("password", ""),
                db       =  self.auth.get("database", ""),
                conv     =  converter)
            
            connection.autocommit(True)
            return connection

        @property
        def server_version(self):
            self._check_connected()
            return self._connection.get_server_info()
        
except ImportError:
    pass
else:
    _server_classes.append(MySQLServer)

class Query(str):
    def substitute(self, rows): pass
    
