# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import re

from boots.api.api import Rows, ResultInfo
from boots.api.nodes.node import NodeGraph, SyncNode, IteratorNode
from boots.api import _, n_

_constructors = {}

def register_constructor(name, constructor):
    """Allow the function constructor to be used from PipedSQL under name.  constructor is a
    function that takes any number of strings as arguments and returns a node."""
    if name in _constructors:
        raise KeyError(_('A constructor with name {0} already exists').format(name))

    _constructors[name] = constructor

def get_constructor(name):
    """Returns the constructor with the specified name."""
    return _constructors[name]

def construct(name, *arguments):
    """Contructs a node using the given name and arguments."""
    if name not in _constructors:
        raise KeyError(_('There is no constructor with the name {0}.').format(name))
    
    return _constructors[name](*arguments)

def register(name):
    """Use this as a decorator to register a constructor."""
    def register_constructor_with_name(constructor):
        register_constructor(name, constructor)
        return constructor

    return register_constructor_with_name

@register('csv_in')
def csv_input_file_node(path):
    """Constructs a node that reads from a csv file and then sends the word to its outputs."""
    input_file = open(path)
    
    def read_iterator():
        for line in input_file:
            for word in line.split(','):
                yield word

    return IteratorNode(read_iterator())

@register('csv_out')
def csv_output_file_node(path):
    """Constructs a node that writes each object it is passed to a file and then sends the object to
    its outputs."""
    class CSVOutput(object):
        def __init__(self):
            self.write_comma = False
            self.output_file = open(path, 'w')
        
        def write(self, result):
            if type(result) is Rows:
                self.output_file.write('\n'.join(','.join(row) for row in result))
                self.output_file.write('\n')
            else:
                if self.write_comma:
                    self.output_file.write(',')
                    
                self.output_file.write(str(result))

                if not isinstance(result, str) or not result.endswith('\n'):
                    self.write_comma = True
                else:
                    self.write_comma = False

            self.output_file.flush()
            
    csv_output = CSVOutput()
    return SyncNode(csv_output.write)

@register('sink')
def sink_node():
    """Constructs a node that ignores all input."""
    return SyncNode(lambda x: None)

@register('sub')
def substitute_node(pattern, replacement, field = None):
    """Constructs a node that replaces any text matching pattern with replacement as if by re.sub.
    field can be used to specify the name of the field to perform substitution against.  The default
    is to substitute against all fields."""

    regexp = re.compile(pattern)

    def sub(result):
        if type(result) is Rows:
            field_indexes = result.field_indexes
            
            if field:
                index = field_indexes[field]
                
                def sub_field(row, replacement):
                    row = list(row)
                    row[index] = replacement
                    return tuple(row)
        
                return Rows((sub_field(row, regexp.sub(replacement, row[index])) for row in result), field_indexes)
            else:
                return Rows((tuple(regexp.sub(replacement, field) for field in row) for row in result), field_indexes)
            
        elif isinstance(result, str):
            return regexp.sub(replacement, result)
        else:
            return result

    return SyncNode(sub)

@register('filter')
def filter_node(pattern, field = None):
    """Constructs a node that allows text to pass through only if it matches pattern.  A row is
    allowed to pass through if at least one of its fields matches pattern.  field can be used to
    specify the name of the field to perform substitution against.  The default is to substitute
    against all fields."""

    regexp = re.compile(pattern)

    def filter(result):
        if type(result) is Rows:
            field_indexes = result.field_indexes
            
            if field:
                index = field_indexes[field]
                return Rows((row for row in result if regexp.match(row[index])), field_indexes)
            else:
                return Rows((row for row in result if any(map(regexp.match, row))), field_indexes)
            
        elif isinstance(result, str):
            if regexp.match(result):
                return result
        else:
            return result

    return SyncNode(filter)
