# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

from __future__ import print_function

import sys
import traceback
from Queue import Queue, Empty
from itertools import chain, tee, repeat, imap, ifilter
from threading import Thread, Event, Lock
from types import GeneratorType

def catch_exceptions(iterable):
    try:
        for x in iterable:
            yield x
    except Exception as e:
        yield e
        
def trace(msg, iterable):
    for x in iterable:
        print(msg, repr(x), file=sys.stderr)
        yield x

def alternate_iterators(*iterables):
    """Iterate over the given iterables, taking turns.

    Takes a list of iterables as input.
    Yields contents of the iterables.
    Removes iterable from the list when iterable is finished.
    Terminates when list is empty."""
    iters = map(iter, iterables)
    while iters:
        for it in iters:
            try:
                yield it.next()
            except StopIteration:
                iters.remove(it)

def flatten(x):
    """Return content objects of an iterable.

    Takes an object as input.
    If an object is iterable, iterate over it. Otherwise, yield the object."""
    if type(x) is GeneratorType:
        for y in x:
            yield y
    else:
        yield x

def imap_flatten(func, iterable):
    """Map a function over the iterable, flattening the result."""
    for x in iterable:
        result = func(x)
        for item in flatten(result):
            yield item

# FIXME: need a thread safe tee here, perhapse "safetee"?

class Status(object):
    def __init__(self, value):
        self.value = value
        
    def __str__(self):
        return str(self.value)
        
def filter_default(value):
    return not isinstance(value, Status) and not isinstance(value, Exception)

def filter_none(value):
    return True

class PrimitiveNode(object):
    def __init__(self, name=None):
        """Constructor of the base node object.

        Initializes the node object.
        Takes the node name, inputs and outputs as parameters.
        """
        self.name = name
        
    @property
    def name(self):
        """Name of the node."""
        # Use a "private" member to avoid name conflicts with Thread
        return self.__name
    
    @name.setter
    def name(self, value):
        """Sets name of the node."""
        self.__name = value
        
    def __repr__(self):
        name = "\"{0}\"".format(self.name) if self.name else "at {0:#x}".format(id(self))
        status_msg = self.status_msg()
        status = " " + status_msg if status_msg else ""
        return "<{classname} {name}{status}>".format(classname = type(self).__name__,
                                                     name      = name,
                                                     status    = status)
    
    def status_msg(self):
        """Returns a status message.

        For the base node object, returns NONE."""
        return None
    
class BaseNode(PrimitiveNode):
    """Base class for the node system."""

    def __init__(self, name=None, inputs=None, outputs=None):
        """Constructor of the base node object.

        Initializes the node object.
        Takes the node name, inputs and outputs as parameters.
        """
        PrimitiveNode.__init__(self, name)
        self._finalized = False
        
        if inputs:
            self.inputs = set(inputs)
        else:
            self.inputs = set()
        
        if outputs:
            self.outputs = set(outputs)
        else:
            self.outputs = set()

    def finalize(self):
        """Finalizes node setup.

        Ensures that a node has the chance to complete its setup without
        its input and output connections changing. In the base node, this
        is not required. Derived classes will need to implement
        _finalize(self)"""
        assert not self._finalized, "Node already finalized."
        self._finalize()
        self._finalized = True    
    
    def insert_before(self, node):
        """Inserts the node before another node.

        Accepts a node as input and prepends the node this method is
        called from to that node."""
        for input in list(self.inputs):
            input.disconnect_output(self)
            input.connect_output(node)
        node.connect_output(self)
        
    def insert_after(self, node):
        """Inserts the node after another node.

        Accepts a node as input and appends the node this method is
        called from to that node."""
        for output in list(self.outputs):
            self.disconnect_output(output)
            node.connect_output(output)
        self.connect_output(node)
        
    def connect_output(self, output_node):
        """Connects output of the node to the input of another node.

        Accepts a target node as parameter and connects this node's output to the
        parameter's input."""
        self.outputs.add(output_node)
        output_node.inputs.add(self)
        
    def disconnect_output(self, output_node):
        """Disconnects the output of this node from the input of another node.

        The opposite of connect_output, taking the same parameters."""
        self.outputs.remove(output_node)
        output_node.inputs.remove(self)
        
    def _iter_inputs(self):
        """Return an iterator over input to the node."""
        return alternate_iterators(*self.inputs)
        
    def _put_outputs(self, data):
        """Send data to all connected output nodes."""
        for output in self.outputs:
            output.put(self, data)
    
    def _finalize(self):
        pass
    
    def start(self):
        pass

class NodeGraph(object):
    """Control and manage a graph of nodes."""
    
    def __init__(self, name=None, nodes=None, auto_process=False):
        """Creates a node graph object.

        Accepts the name of the graph, the content nodes, and whether or not to automatically
        process the graph."""
        self.name = name
        self.nodes = set()
        # TODO: auto-connect nodes in node graph to the input and output links
        self.input_link = LinkNode("{0}-input".format(name))
        self.output_link = LinkNode("{0}-output".format(name), mortal=True)
        self.names = {}
        self.auto_process = auto_process
        self._finalized = False
        self.async = False
        if nodes:
            self.add_connected(*nodes)
    
    @staticmethod
    def from_sequence(node_seq):
        """Connects a sequence of nodes.

        Connects a sequence of nodes, with branches for sub-sequences.
        For example, an input (1,2,3) will return -1-2-3-; an input of
        ((1,2,3),a,b,c) will return -1-2-3-
                                     |
                                      -a-b-c
        """
        def get_first(obj):
            try:
                return obj[0]
            except TypeError:
                return obj
        
        def add_seq(graph, seq):
            for index, node in enumerate(seq):
                if not isinstance(node, BaseNode):
                    add_seq(graph, node)
                    node = node[0]
                    
                if index > 0:
                    node.inputs.add(get_first(seq[index - 1]))
                if index < len(seq) - 1:
                    node.outputs.add(get_first(seq[index + 1]))
                graph.add(node)
        
        node_graph = NodeGraph()
        add_seq(node_graph, node_seq)
        return node_graph
    
    def edges(self):
        """Return the set of edges in the node graph."""
        input_edges = set()
        output_edges = set()
        for node in chain(self, [self.input_link, self.output_link]):
            for input in node.inputs:
                input_edges.add((input, node))
            for output in node.outputs:
                output_edges.add((node, output))
        assert input_edges == output_edges, "Graph integrity check failed."
        return output_edges
        
    def add(self, *nodes, **flags):
        """Add individual nodes to the collection.

        Accepts a list of nodes to be added to the collection."""
        self._add(False, nodes)
        
        if flags.get("input", False):
            for node in nodes:
                self.input_link.connect_output(node)
        
        if flags.get("output", False):
            for node in nodes:
                node.connect_output(self.output_link)
        
        return self
    
    def add_input(self, *nodes):
        """Adds nodes to the input link.

        Accepts a list of nodes and adds them to the input_link of this
        graph."""
        self.add(*nodes, input=True)
        return self
        
    def add_output(self, *nodes):
        """Adds nodes to the output link.

        Accepts a list of nodes and adds them to the output_link of this
        graph."""
        self.add(*nodes, output=True)
        return self
        
    def add_connected(self, *nodes):
        """Add all connected nodes to the collection."""
        self._add(True, nodes)
        return self

    def _add(self, traverse, nodes):
        if traverse:
            node_iter = self._iter_nodes(nodes)
        else:
            node_iter = iter(nodes)
            
        for node in node_iter:
            self.nodes.add(node)
            if node.name:
                self.names[node.name] = node
    
    def link_output(self, output_graph):
        """Link output link of this node graph to input of another."""
        self.output_link.link_output(output_graph.input_link)
        
    def unlink_output(self, output_graph):
        """Unlink output link of this node graph from input of another."""
        self.output_link.unlink_output(output_graph.input_link)
        
    def __getitem__(self, name):
        return self.names[name]
    
    def __iter__(self):
        return iter(self.nodes)
    
    @classmethod
    def _iter_nodes(cls, nodes, seen=None):
        """Iterate over a depth first traversal of the graph for each given 
        node, skipping already seen nodes.
        
        """
        if not seen:
            seen = set()
        
        for node in nodes:
            if node not in seen:
                seen.add(node)
                yield node
                for traversed_node in cls._iter_nodes(chain(node.inputs, node.outputs), seen):
                    yield traversed_node

    def _process(self):        
        # FIXME: This code is ugly and difficult to read.
        for node in list(self.nodes):
            # Ensure that all sync "tails" end with async nodes (threads)
            queue_outputs = set(output for output in node.outputs if isinstance(output, QueueNode))
            if isinstance(node, SyncNode) and (not node.outputs or queue_outputs):
                # Add an async node with the identity function so the SyncNode(s) run in a thread.
                async_node = AsyncNode()
                
                # We want to insert the AsyncNode in between the SyncNode and any queues it is connected to.
                for queue in queue_outputs:
                    node.disconnect_output(queue)
                    async_node.connect_output(queue)
                    
                node.connect_output(*self.add(async_node))
                
            # Add queues at merge points and input from non-sync nodes
            if len(node.inputs) > 1 or (len(node.inputs) == 1
                                        and not isinstance(tuple(node.inputs)[0], (SyncNode, QueueNode))
                                        and not isinstance(node, QueueNode)):
                node.insert_before(*self.add(QueueNode()))
    
    def finalize(self):
        """Finalize all nodes in node graph.

        Also processes graph if auto_process is set
        """
        assert not self._finalized, "Graph already finalized."
        
        if self.auto_process:
            self._process()       

        # Notify all nodes that their connections are final.
        for node in self.nodes:
            node.finalize()
            
        self.input_link.finalize()
        self.output_link.finalize()
                
        self._finalized = True
        
    def start(self):
        """Start data processing in the node graph."""
        self.finalize()
        for node in self:
            node.start()
            if isinstance(node, AsyncNode):
                self.async = True
            
    def put(self, origin, data):
        """Put data on input link."""
        self.input_link.put(origin, data)
        
    def emit(self, data):
        self.output_link.emit(data)
    
    def run(self):
        """Output data."""
        if not self.async:
            self.output_link.run()

class ActionNode(BaseNode):
    def __init__(self, action, name=None, inputs=None, outputs=None, filter=None, output_none=False):
        BaseNode.__init__(self, name, inputs, outputs)
        self.filter = filter or filter_default
        self.output_none = output_none
        self.action = action
    
    def _do_action(self, data):
        if self.filter(data):
            return self.action(data)
        else:
            return data
    
    def _run(self):
        for data in self._iter_inputs():
            try:
                self.emit(self._do_action(data))
            except Exception as e:
                self.put_outputs(e)
        self._put_outputs(StopIteration)
        
    def emit(self, data):
        for result in flatten(data):
            if result is not None or self.output_none:
                self._put_outputs(result)
        
    def _finalize(self):
        pass
    
    def start(self):
        pass
    
class QueueNode(BaseNode):
    """Thread-safe queue that buffers pushed data for pull access."""
    
    def __init__(self, name=None, inputs=None, outputs=None, block=True):
        BaseNode.__init__(self, name, inputs, outputs)
        self.block = block
        self._queue = Queue()
        self._unfinished_inputs = None
    
    def _finalize(self):
        self._unfinished_inputs = len(self.inputs)
        
    def put(self, origin, data):
        self._queue.put(data)
    
    def __iter__(self):
        while self._unfinished_inputs or not self.inputs:
            try:
                data = self._queue.get(self.block)
            except Empty:
                data = None
            
            if data is StopIteration and self.inputs:
                self._unfinished_inputs -= 1
                if self._unfinished_inputs:
                    break
            else:
                yield data
            
class BaseLinkNode(object):
    def __init__(self, include_origin):
        """Initializes link node."""
        self.input_links = set()
        self.output_links = set()
        self.input_lock = Lock()
        self.output_lock = Lock()
        self.include_origin = include_origin
    
    def link_output(self, output_link):
        """Links own output to another unit's input link."""
        with self.output_lock:
            with output_link.input_lock:
                self.output_links.add(output_link)
                output_link.input_links.add(self)
        
    def unlink_output(self, output_link):
        """Unlinks own output from another unit's input link."""
        with self.output_lock:
            with output_link.input_lock:
                self.output_links.remove(output_link)
                output_link.input_links.remove(self)
                
    def _result(self, origin, data):
        if self.include_origin:
            return (origin, data)
        else:
            return data
        
    def emit(self, data):
        output_links = None
        with self.output_lock:
            output_links = self.output_links.copy()
        
        for output_link in output_links:
            output_link.put(self, data)
            
    def run(self):
        for data in self:
            self.emit(data)
            if data is StopIteration:
                break
            
class LinkNode(BaseNode, BaseLinkNode):
    """Link node to connect to nodes in a dynamic, thread-safe manner."""

    def __init__(self, name=None, inputs=None, outputs=None, include_origin=False, mortal=False):
        """Initializes link node."""
        BaseNode.__init__(self, name, inputs, outputs)
        BaseLinkNode.__init__(self, include_origin)
        self.mortal = mortal
        
    def _finalize(self):
        self._input_iter = imap(lambda x: self._result(self, x), self._iter_inputs())
    
    def emit(self, data):
        self._put_outputs(data)
        BaseLinkNode.emit(self, data)
    
    def put(self, origin, data):
        """Push data through the link."""
        if not (data is StopIteration and not self.mortal):
            self.emit(self._result(origin, data))
        
    def _iter_links(self):
        while True:
            input_links = None
            with self.input_lock:
                input_links = self.input_links.copy()

            if input_links:
                for link in input_links:
                    try:
                        yield self._result(link, iter(link).next())  
                    except StopIteration:
                        link.unlink_output(self)
            else:
                if self.mortal:
                    return
                else:
                    yield None, None
        
    def __iter__(self):
        return alternate_iterators(self._input_iter, self._iter_links())

class LinkStackNode(PrimitiveNode, BaseLinkNode):
    """Link node to read input from a stack of nodes, popping off the stack
    when a node finishes."""

    def __init__(self, name=None, include_origin=False):
        """Initializes link node."""
        PrimitiveNode.__init__(self, name)
        BaseLinkNode.__init__(self, include_origin)
        self.input_stack = list()
        
    def push_input(self, input_link):
        with self.input_lock:
            with input_link.output_lock:
                self.input_stack.append(input_link)
                input_link.output_links.add(self)
        
    def pop_input(self):
        with self.input_lock:
            input_link = self.input_stack.pop()
            with input_link.output_lock:
                input_link.output_links.remove(self) 

    def __iter__(self):
        while True:
            with self.input_lock:
                top_input = self.input_stack[-1] if self.input_stack else None
                    
            if top_input:
                try:
                    yield self._result(top_input, iter(top_input).next())
                except StopIteration:
                    self.pop_input()
            else:
                return
                    
class SyncNode(ActionNode):
    """Synchronous lazy pull-based node."""   
    def _finalize(self):
        if self.inputs:
            result_iter = imap_flatten(self._do_action, self._iter_inputs())
        else:
            result_iter = self.action
            
        if not self.output_none:
            result_iter = ifilter(lambda x: x is not None, result_iter)
            
        self._result_iter = result_iter
        self._tees = list(tee(catch_exceptions(self._result_iter), len(self.outputs)))
    
    def put(self, origin, data):
        if data is StopIteration:
            self._put_outputs(StopIteration)
        else:
            self.emit(self._do_action(data))
        
    def __iter__(self):
        return self._tees.pop()
        
    def start(self):
        pass
    
    def run(self):
        self._run()

class IteratorNode(SyncNode):
    def __init__(self, value, name=None, inputs=None, outputs=None, filter=None):
        self.iterator = iter(value)
        SyncNode.__init__(self, self.iterator, name, inputs, outputs, filter)

class AsyncNode(ActionNode, Thread):
    """Asynchronous push-based node. Runs in a thread."""
    
    def __init__(self, action=None, name=None, inputs=None, outputs=None, filter=None, output_none=False, profile=False):
        Thread.__init__(self)
        ActionNode.__init__(self, action, name, inputs, outputs, filter, output_none)
        if action:
            self.action = action
        else:
            self.action = lambda x:x 
        self.profile = profile
        self.finished = Event()
    
    def status_msg(self):
        if self.is_alive():
            return "running"
        else:
            return "stopped"
    
    def start(self):
        Thread.start(self)
    
    def run(self):
        if not self.profile:
            self._run()
        else:
            import cProfile
            cProfile.runctx("self._run()", globals(), locals(), self.profile)
    
    def _run(self):
        ActionNode._run(self)
        self.finished.set()

def test():
    def wait(x):
        import time
        print("waiting: {0}".format(x))
        time.sleep(1)
        return x
    
    def print_it(foo):
        print("printing: {0}".format(foo))
        return foo
    
    nodes = NodeGraph.from_sequence([SyncNode(lambda x:xrange(10), name="count"),
                                     AsyncNode(wait, name="wait"),
                                     SyncNode(print_it, name="print"),
                                     QueueNode(name="out")])
    nodes.start()
    print(nodes.edges())
    for result in nodes["out"]:
        print(result)
        
if __name__ == "__main__":
    test()
