# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides a class that builds a boots client configuration. The
configuration is built using a set of defaults, an rc file, and command line
options/arguments."""

import optparse
import os, os.path
import info

from boots.app import _, n_

def find_executable(name):
    # Search algorithm adapted from os._execvpe 
    head, tail = os.path.split(name)
    if head and os.path.exists(name):
        return name
    else:
        if 'PATH' in os.environ:
            envpath = os.environ['PATH']
        else:
            envpath = os.defpath
        PATH = envpath.split(os.pathsep)
        for dir in PATH:
            fullname = os.path.join(dir, name)
            if os.path.exists(fullname):
                return fullname

def get_auto_pager():
    if find_executable("less"):
        return "less -XF"
    elif find_executable("more"):
        return "more"    

class ClientConfig(object):
    """Class used to read client configuration from files and the cli."""

    def __init__(self, conf_file_loc = None):
        """Initialize a configuration. Optionally takes a path to the rc file.

        This method performs no IO and after initializaion you must call load
        to populate a config."""
        # Specify default values for a client configuration.
        self._defaults = {"command":            None,
                          "database":           None,
                          "script":             None,
                          "lingo":              "sql",
                          "rcfile":             info.RCFILE,
                          "host":               "localhost",
                          "port":               3306,
                          "username":           os.getenv("USER"),
                          "password":           False,
                          "prompt1":            "> ",
                          "prompt2":            "+ ",
                          "pager":              None,
                          "pager_command":      None,
                          "terminating_char":   ";",
                          "history_length":     100,
                          "history_file":       os.path.expanduser("~/.boots_history")}
        self._dict = self._defaults.copy()

        self.conf_file = conf_file_loc

        # Build cli option parser.
        version_string = "{name} version {version}".format(name = info.NAME,
                                                           version = info.VERSION)
        self._cli_parser = optparse.OptionParser(version = version_string);
        self._cli_parser.add_option("-c", "--command",
                                    action = "store",
                                    type = "string",
                                    dest = "command",
                                    help = _("Execute command and exit"))
        self._cli_parser.add_option("-D", "--database",
                                    action = "store",
                                    type = "string",
                                    dest = "database",
                                    help = _("Use database"))
        self._cli_parser.add_option("-f", "--script",
                                    action = "store",
                                    type = "string",
                                    dest = "script",
                                    help = _("Execute commands from file and exit"))
        self._cli_parser.add_option('-l', '--lingo',
                                    action = 'store',
                                    type = 'string',
                                    dest = 'lingo',
                                    help = _('Set the lingo to use'))
        self._cli_parser.add_option("-X", "--norc",
                                    action = "store_true",
                                    dest = "norc",
                                    help = _("Do not read user configuration file"))
        self._cli_parser.add_option("-r", "--rc",
                                    action = "store",
                                    type = "string",
                                    dest = "rcfile",
                                    help = _("Filename of user configuration file"))
        self._cli_parser.add_option("-H", "--host",
                                    action = "store",
                                    type = "string",
                                    dest = "host",
                                    help = _("Connect to host"))
        self._cli_parser.add_option("-p", "--port",
                                    action = "store",
                                    type = "int",
                                    dest = "port",
                                    help = _("Use port number"))
        self._cli_parser.add_option("-u", "--user",
                                    action = "store",
                                    type = "string",
                                    dest = "username",
                                    help = _("Login with username"))
        self._cli_parser.add_option("-P", "--password",
                                    action = "store",
                                    type = "string",
                                    dest = "password",
                                    help = _("Connect using password. If none is given, query for password"))
        self._cli_parser.add_option("--pager",
                                    action = "store",
                                    type = "string",
                                    dest = "pager",
                                    help = _("Pipe query results to the specified pager."))
        self._cli_parser.add_option("-t", "--terminatingchar",
                                    action = "store",
                                    type = "string",
                                    dest = "terminating_char",
                                    help = _("Specify the SQL statement terminating character. Default is ';'."))
        self._cli_parser.add_option("-F", "--historyfile",
                                    action = "store",
                                    type = "string",
                                    dest = "history_file",
                                    help = _("Specify file to save history to"))
        self._cli_parser.add_option("-L", "--historylength",
                                    action = "store",
                                    type = "int",
                                    dest = "history_length",
                                    help = _("Specify max history file length"))
 
    def __getitem__(self, key):
        """Allows one to use the [] operator to get config items."""
        return self._dict[key]

    def __setitem__(self, key, value):
        """Allows one to use the [] to set config items."""
        self._dict[key] = value

    @property
    def options(self):
        """Give entire configuration as a dictionary."""
        return self._dict

    def load(self):
        """Load a configuration from the cli and rc file.

        Options specified in rc file overwrite defaults and cli options
        overwrite rc options."""
        from_cli = self.get_cli_conf()
        
        if not self.conf_file:
            if "rcfile" in from_cli:
                self.conf_file = from_cli["rcfile"]
            else:
                self.conf_file = self._defaults["rcfile"]
        from_file = self.get_file_conf(self.conf_file)
        
        self._dict.update(from_file)
        self._dict.update(from_cli)
        self._interpret_options()
 
    def _interpret_options(self):
        # Interpret options
        if self["pager"] == "auto":
            self["pager_command"] = get_auto_pager()
        elif self["pager"]:
            self["pager_command"] = self["pager"]

    def get_file_conf(self, filepath):
        """Read a configuration from the specified file. Return a dict."""
        file_dict = {}
        try:
            execfile(filepath, {}, file_dict)
        except IOError:
            if filepath != info.RCFILE:
                # Only print an error if a non-default rc file was specified.
                print(_("rcfile {0} not found").format(filepath))
        except SyntaxError:
            print(_("rcfile {0} contains a syntax error").format(filepath))
        except Exception as e:
            print(_("rcfile {0} contains an error: {1}").format(filepath, e))
        return file_dict

    def get_cli_conf(self):
        """Read a configuration from the cli. Return a dict."""
        (options, args) = self._cli_parser.parse_args()
        tmp_options = vars(options)
        new_options = {}
        for key, value in tmp_options.items():
            if value != None:
                new_options[key] = value
        self._args = args
        return new_options

    def update(self, pairs_to_add):
        """Update configuration with provided dictionary."""
        self._dict.update(pairs_to_add)
