# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides the boots Console class. This class is the core
driver of any boots client."""

import sys
import signal

from boots.api import api
from boots.api.errors import BootsError, BootsWarning, ConnectionError
from boots.api.nodes.node import NodeGraph, SyncNode, IteratorNode, LinkStackNode
from boots.lib.ui.plain import PlainUI
from boots.lib.ui.generic import ScriptDriver, StringDriver, StdoutPresenter
from boots.lib.ui.components.help import HelpTopic, QueryIndexTopic, CommandHelpTopic
from boots.lib.ui.components.metacommands import MetaCommandManager, MetaCommands, parse_metacommand
from boots.lib.hook import Hooks
from boots.lib.lingos.lingo import lingo_registry
from boots.lib.lingos import bash_external
from boots.lib.lingos import python
from boots.lib.lingos import piped_sql
from boots.lib.lingos import sql
from boots.lib import _, n_
try:
    from boots.lib.lingos.lisp import lisp
except ImportError:
    # Allow the lisp lingo import to fail if Ply is not installed.
    pass

# Utility function.
def iterable(obj):
    """Returns True if obj is iterable; otherwise, False is returned."""
    try:
        iter(obj)
        return True
    except TypeError:
        return False
    
class ConsoleError(BootsError):
    """Errors to be displayed on the console."""
    pass

class InternalError(BootsError):
    """Errors to be displayed on the console."""
    def __init__(self, exception):
        self.exception = exception
        
    def __str__(self):
        return _("Boots encountered an internal error: {0}").format(self.exception)

class Console(object):    
    def __init__(self, config, welcome_msg=None):
        self.config = config
        self.welcome_msg = welcome_msg
        
        self.hooks = Hooks("load", "unload")

    def _setup(self):
        # Setup signal handling as the first thing we do.
        signal.signal(signal.SIGTERM, self.handle_signal)
        
        self._init_help()
        
        # Set up metacommands
        self.metacommandmanager = MetaCommandManager()
        self.metacommands = MetaCommands({
            '\connect':    self.meta_connect,
            '\disconnect': self.meta_disconnect,
            '\quit':       self.quit,
            '\help':       self.show_help,
            '\h':          self.show_help,
            'help':        self.show_help,
            'source':      self.meta_source,
            '\.':          self.meta_source,})
        self.metacommandmanager.add(self.metacommands)
        
        self.servers = []
        
        # Initialize lingos 
        self.lingos = lingo_registry.instantiate(self)
        for name, lingo in self.lingos.iteritems():
            if hasattr(lingo, "help"):
                self.help["lingos"][name] = lingo.help

        # Set up UI components
        self.ui = PlainUI(self)
        self.presenter = self.ui
        if self.config["script"]:
            self.driver = ScriptDriver(self, self.config["script"], self.config["lingo"])
        elif self.config["command"]:
            self.driver = StringDriver(self, self.config["command"], self.config["lingo"])
        else:
            self.driver = self.ui
        
        if hasattr(self.ui, "metacommands"):
            self.metacommandmanager.add(self.ui.metacommands)
            
        if hasattr(self.ui, "help"):
            self.help["ui"] = self.ui.help
        
        # Set up core node graphs
        self.graph = NodeGraph("Console")
        self.graph.add(SyncNode(self._interact, name="Console-interact"), input=True, output=True)
        self.graph.input_link.mortal = True
        self.graph.input_link.include_origin = True

        self.driver_stack = LinkStackNode()        
        self.driver_stack.link_output(self.graph.input_link)
        self.driver_stack.push_input(self.driver.driver_graph.output_link)
        
        self.graph.link_output(self.presenter.presenter_graph)
        self.presenter.presenter_graph.link_output(self.graph)
        
        self.presenter.presenter_graph.start()
        self.driver.driver_graph.start()
            
        self.hooks["load"].call()

    def _init_help(self):
        self.help = HelpTopic("boots", _("Boots Help"), _("Help"),
            description = _("Boots is a flexible, extensible, and multilingual shell for working with databases. To read more about a subtopic, use `help <topic>`."))
        
        self.help.add("commands", _("Commands"), _("List of Boots commands."),
            query = ("commands",),
            kind = QueryIndexTopic)
        
        self.help["commands"].add("\connect", _("connect"), _("Connect to a server."),
            usage       = "\connect [host] [port] [database]",
            description = _("Connect to host:port, using the database specified."),
            kind = CommandHelpTopic)        
        
        self.help["commands"].add("\disconnect", _("disconnect"), _("Disconnect from a server."),
            usage       = "\disconnect [host] [port]",
            description = _("Disconnect from the specified host."),
            kind = CommandHelpTopic)
        
        self.help["commands"].add("\help", _("help"), _("Show help for a topic."),
            usage       = "\help topic",
            description = _("Show help for \"topic\"."),
            aliases = ["\h", "help"],
            kind = CommandHelpTopic)
        
        self.help["commands"].add("\.", _("source"), _("Load and execute a file."),
            usage       = "\source filename",
            description = _("Load and execute a file containing database queries."),
            aliases = ["\source"],
            kind = CommandHelpTopic)

        self.help["commands"].add("\quit", _("quit"), _("Quit boots."),
            usage       = "\quit",
            description = _("Quit boots."),
            kind = CommandHelpTopic)
            
        self.help.add("lingos", _("Lingos"), _("Documentation for loaded lingos."),
            description = _("A lingo is a command language usable within Boots."))

    def input_complete(self, command, lingo):
        if self.metacommands.get_metacommand(command) is not None:
            return command 
        elif lingo in self.lingos:
            return self.lingos[lingo].input_complete(command)
        else:
            return command
                      
    def run(self, command, lingo="sql"):
        if not lingo in self.lingos:
            raise ConsoleError(_("Invalid lingo \"{0}\" specified.").format(lingo))
        
        result_data = self.lingos[lingo].execute(command)
        result = self.lingos[lingo].present(result_data)
        if isinstance(result, NodeGraph):
            result_graph = result
        else:
            result_graph = NodeGraph() 
            if isinstance(result, (str, Exception)) or not iterable(result):
                result = iter([result])
            result_graph.add_output(IteratorNode(result))

        result_graph.link_output(self.presenter.presenter_graph)
        result_graph.start()
        result_graph.run()
        return result_data

    def connect(self, host, port, database):
        try:
            server = api.DrizzleServer(host, port, {"database": database})
            server.connect()
        except ConnectionError:
            server = None
            
        if server and server.is_connected:
            self.servers.append(server)
            return True
        else:
            raise ConsoleError(_("Could not connect to {0}:{1}").format(host, port))

    def disconnect(self, host, port):
        found = False
        for server in self.servers:
            if not host or server.host == host and not port or server.port == port:
                server.disconnect()
                self.servers.remove(server)
                found = True
        return found
    
    def _load_servers(self):
        self.connect(self.config["host"], self.config["port"], self.config["database"])
    
    def _welcome_message(self):
        if self.welcome_msg and self.driver.is_interactive:
            server_lines = []
            for server in self.servers:
                line_format = _("{0.hostname}:{0.port} (server v{0.server_version})")
                server_lines.append(line_format.format(server))
            server_status = "\n".join(server_lines)
            
            welcome = self.welcome_msg.format(server_status=server_status)
            self.graph.emit(welcome)
    
    def _error(self, error):
        if not isinstance(error, (BootsError, BootsWarning)):
            error = InternalError(error)
        return error
    
    def main(self):
        self._setup()
        
        try:
            self._load_servers()
        except ConsoleError as e:
            self.graph.emit(e)
        
        self._welcome_message()
        
        self.graph.start()
        self.driver_stack.run()
        
        self.quit()
        
    def _interact(self, data):
        origin, data = data
        if origin is self.driver_stack:
            if isinstance(data, Exception):
                return self._error(data)
            else:
                command, lingo = data
                try:
                    if not self.metacommands.execute(command):
                        self.run(command, lingo)
                except Exception as e:
                    return self._error(e)
                    
        elif origin is self.presenter.presenter_graph.output_link:
            if isinstance(data, Exception) and not self.driver.is_interactive:
                self.quit(exit_code=1)
    
    def handle_signal(self, signum, frame):
        if signum == signal.SIGTERM:
            self.quit()
    
    def quit(self, exit_code=0):
        if self.driver.is_interactive:
            self.graph.emit(_("Boots quit."))
            
        self.hooks["unload"].call()
        self.disconnect(None, None)
        sys.exit(exit_code)
        
    def show_help(self, *query):
        results = list(self.help.search(query))
        if results:
            # Do not disambiguate exactly matching queries
            for topic, path in results:
                if path == query:
                    results = [(topic, path)]
                    break
                
            if len(results) == 1:
                topic, path = results[0]
                self.presenter.present(topic.format(include_index=True))
            else:
                paths = ", ".join("\"{0}\"".format(" ".join(path)) for topic, path in results)
                self.presenter.present(_("Did you mean: {0}?").format(paths))
        else:
            self.presenter.present(_("No help found for \"{0}\".").format(" ".join(query)))

    def meta_connect(self, host=None, port=None, database=None):
        if host is None: 
            host = self.config["host"]
            
        if port is None:
            port = self.config["port"]
        else:
            port = int(port)
            
        if database is None:
            database = self.config["database"]

        self.connect(host, port, database)

    def meta_disconnect(self, host=None, port=None):
        if host is not None:
            host = str(host)
        if port is not None:
            port = int(port)

        self.disconnect(host, port)

    def meta_source(self, *file_list):
        for filepath in file_list:
            try:
                script_driver = ScriptDriver(self, filepath, self.config["lingo"])
            except IOError as e:
                raise ConsoleError(_("Could not open file {0}").format(filepath))
            
            self.driver_stack.push_input(script_driver.driver_graph.output_link)
            script_driver.driver_graph.start()
