# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import sys

from boots.lib import _, n_

class LingoRegistry(object):
    def __init__(self):
        self._lingos = {}
        
    def register(self, name, interpreter):
        """Registers a lingo with the given name and interpreter.  Raises a KeyError if a lingo with
        name is already registered."""
        if name in self._lingos:
            raise KeyError(_('A lingo with name {0} already exists').format(name))
    
        self._lingos[name] = interpreter
    
    def __getitem__(self, name):
        """Return the interpreter for the lingo of the given name.  If no such lingo exists, return
        None."""
        return self._lingos[name]
    
    def __iter__(self):
        return self._lingos.iteritems()
    
    def instantiate(self, *args, **kwargs):
        instances = {} 
        for name, lingo_cls in self:
            instances[name] = lingo_cls(*args, **kwargs)
        return instances

lingo_registry = LingoRegistry()
register = lingo_registry.register

# The base class for all Interpreters.
class Interpreter(object):
    # FIXME: Once multiple servers are supported, servers should default to the current list of
    # servers and the servers argument here should be removed.
    def __init__(self, console, stdin = sys.stdin, stdout = sys.stdout):
        """Create a new Interpreter object.  stdin will be used as the standard input within the
        lingo and may be either a file or a callable.  If stdin is a callable, each invocation
        should take no arguments and return the next line of input.  Newlines should be retained.
        stdout should be either a file or a callable.  If stdout is a callable it should take one
        string argument.  servers is a list of servers to execute commands on."""
        self.console = console

        if isinstance(stdin, file):
            self.stdin = stdin.readline
        else:
            self.stdin = stdin

        if isinstance(stdout, file):
            self.stdout = stdout.write
        else:
            self.stdout = stdout

    def input_complete(self, command):
        """Indicates if command is a complete command for the given lingo by returning the
        command if it is complete and None otherwise."""
        return command
    
    def read(self, command):
        """Convert the string command to the format used to represent code.  This could be an
        abstract syntax tree or even plain text.  The result must be an object that can be passed to
        evaluate."""
        return command

    def evaluate(self, code):
        """Evaluate code; return the result of evaluating code in the object system for this lingo.
        code should be an object returned by read."""
        pass

    def present(self, obj):
        """Convert obj to a presentable object.  obj is an object in the object system for this lingo."""
        return obj
    
    def python(self, obj):
        """Convert obj to the appropriate python type.  obj is an object in the object system for
        this lingo."""
        return obj
  
    def execute(self, command):
        """Equivalent to self.python(self.evaluate(self.read(command)))."""
        return self.python(self.evaluate(self.read(command)))
