# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

# A simple lisp interpreter written in Python.

# To Do:
# Add support for:
# defun
# defmacro
# let
# do

import sys

if __name__ == '__main__':
    sys.path.append('../../../../')

import re
import readline

from boots.lib.lingos import lingo
from boots.lib.lingos.lisp import builtins
from boots.lib.lingos.lisp import lexer
from boots.lib.lingos.lisp import objects
from boots.lib.lingos.lisp import parser
from boots.lib import _, n_


def input_complete(string):
    """Returns string if string is a complete lisp expression.  Otherwise, None is returned."""
    # Remove strings as they may contain parentheses.
    tmp = re.sub(lexer.t_STRING, '', string)
    left_parens = tmp.count('(')
    right_parens = tmp.count(')')
    paren_level = left_parens - right_parens

    if paren_level < 0:
        raise SyntaxError(_('unmatched closing ")"'))
    elif paren_level > 0:
        return None
    else:
        return string

def read_string(string):
    """Reads string and returns a lisp object."""
    return parser.parser.parse(string)

def read(stream = sys.stdin):
    """Reads a lisp object from stream.  stream may be either a file or a callable in which case each call to stream should
    return the next line."""
    paren_level = 0
    lines = ''
    
    if isinstance(stream, file):
        stream = stream.readline
    elif not callable(stream):
        raise TypeError(_('{0} is not a stream or a callable').format(stream))
    # Read lines until all open lisp expressions have been closed.
    while True:
        line = stream()
        # Put a space between arguments separated by a newline to prevent concatenation.
        lines += ' ' + line

        if input_complete(lines):
            break
        
    return read_string(lines)

def evaluate(expression, environment):
    """Evaluates expression in environment and returns the result."""
    return expression.evaluate(environment)

class LispInterpreter(lingo.Interpreter):
    def __init__(self, *arguments, **keywords):
        super(LispInterpreter, self).__init__(*arguments, **keywords)
        self.global_environment = objects.Environment()
        builtins.initialize(self.global_environment, self.stdin, self.stdout)

    def input_complete(self, command):
        return input_complete(command)
    
    def read(self, command):
        if command.strip():
            return read_string(command)

    def evaluate(self, code):
        if code != None:
            return evaluate(code, self.global_environment)

lingo.register('lisp', LispInterpreter)

def main():
    try:
        global_environment = objects.Environment()
        builtins.initialize(global_environment)
        
        while True:
            try:
                print(evaluate(read(lambda: raw_input('> ')), global_environment))
            except SyntaxError as e:
                print(e)
            
    except EOFError:
        pass

if __name__ == '__main__':
    main()
