# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import re

from boots.api.nodes.node import NodeGraph, SyncNode, IteratorNode
from boots.api.constructors import construct, csv_output_file_node, sink_node
from boots.lib.ui.components.help import HelpTopic, CommandHelpTopic
from boots.lib.lingos import lingo
from boots.lib.lingos import sql
from boots.lib import _, n_

# Set this to another node constructor to change the default constructor used for output files.
output_file_node = csv_output_file_node

# This regexp matches a clause in the pipeline.  Use the input_file group for the name of the input
# file, the constructor_call group as the text to be matched against constructor_call_re and
# the output_file group for the name of the output file.
clause_re = re.compile('^(?P<constructor_call>[^<>|]+)([ ]*>[ ]*(?P<output_file>[^>|]+))?$')
# This regexp matches a function call syntax.  Use the name group to access the name of the function
# and the arguments group to access the list of arguments (as a string).
constructor_call_re = re.compile('^(?P<name>[a-zA-Z_]+)\((?P<arguments>([^,\(\)]+,)*[^,\(\)]+)?\)$')
# Matches a string of the form 'abc' or "abc".
string_re = re.compile('^(?P<text>(\'[^\']*\')|("[^"]*"))')

class PipedSQLInterpreter(sql.SQLInterpreter):
    help = HelpTopic("pipedsql", _("PipedSQL Lingo"), _("Enhanced SQL query language supporting pipes."),
        description = _("PipedSQL is an enhanced SQL-like query language for Boots, allowing data to be redirected through pipes. Normal SQL queries can be run in PipedSQL without modification, with the option of piping the results to a chain of PipedSQL objects.\n"
                        "\nExample: SELECT * FROM table | csv_out(\"result.csv\")"),
        see_also    = ["pipedsql objects"])
    
    help.add("objects", _("PipedSQL Objects"), _("Objects to pipe to in PipedSQL."))
    
    help["objects"].add("csv_out", "csv_out", _("Outputs piped data to a CSV file."),
        usage       = "csv_out(path)",
        description = _("Comma separated value data output. Takes one argument, a file path to write to."),
        kind = CommandHelpTopic)
    
    help["objects"].add("filter",  "filter", _("Returns only indicated text."),
        usage       = "filter(pattern[, column])",
        description = _("Takes two arguments. Filters out all text that does not match the first argument. The second argument is an optional name of a column to filter."),
        kind = CommandHelpTopic)
    
    help["objects"].add("csv_in", "csv_in", _("Inputs piped data from a CSV file."),
        usage       = "csv_in(path)",
        description = _("Comma separated value data input. Takes one argument, a file path to read from."),
        kind = CommandHelpTopic)
    
    help["objects"].add("sub", "sub", _("Replaces indicated text with substitution text."),
        usage       = "sub(pattern, replacement[, column])",
        description = _("Takes three arguments. Replaces the first argument with the text of the second. The third argument is an optional name of of a column to limit replacements to."),
        kind = CommandHelpTopic)
    
    def __init__(self, *args, **kwargs):
        super(PipedSQLInterpreter, self).__init__(*args, **kwargs)
        # FIXME: Make PipedSQL work across multiple lines.  This requires more than just removing
        # the following line so don't do that!
        self.terminating_char = '\n'

    def _parse_clause(self, clause_string):
        match = clause_re.match(clause_string)

        if not match:
            raise SyntaxError(_('{0} is not in the correct function call syntax').format(clause_string))

        constructor_call = match.group('constructor_call').strip()
        output_file = match.group('output_file')
        return constructor_call, output_file
        
    def _parse_constructor_call(self, constructor_string):
        # I prefer to use re instead of parser here because it allows for more control.
        match = constructor_call_re.match(constructor_string)

        if not match:
            raise SyntaxError(_('{0} is not in the correct function call syntax').format(constructor_string))
        
        name = match.group('name')
        arguments = match.group('arguments')

        if arguments:
            arguments = [string_re.match(argument.strip()).group('text')[1:-1] for argument in arguments.split(',')]
        else:
            arguments = []
        
        return name, arguments
        
    def evaluate(self, code):
        """Contruct a node graph that will implement the pipeline given by code."""
        if code.strip():
            nodes = NodeGraph()
            first_node = True
            last_node = None

            for clause in code.split('|'):
                sql = False
                # Each clause in the pipeline may contain output redirection.  SQL is allowed only
                # for the first node in the chain.
                try:
                    constructor_call, output_file = self._parse_clause(clause)
                except SyntaxError as e:
                    print(e)
                    print _('error1')
                    return [] # Return an empty list to avoid printing None.

                try:
                    constructor, arguments = self._parse_constructor_call(constructor_call)
                except SyntaxError:
                    # If it wasn't a constructor call, assume it was SQL.
                    sql = True

                # SQL is only allowed for the first node.
                if sql:
                    if not first_node:
                        raise SyntaxError(_('SQL is only permitted for the first clause'))
                    
                    # FIXME: Add support for multiple servers here.  Currently only the first server
                    # is used.
                    cmd_queue = IteratorNode([constructor_call])
                    nodes.add(cmd_queue)
                    server = self.console.servers[0]
                    server_node = SyncNode(server.execute)
                    nodes.add(server_node)
                    cmd_queue.connect_output(server_node)
                    node = server_node
                else:
                    try:
                        node = construct(constructor, *arguments)
                        nodes.add(node)
                        
                        if last_node:
                            last_node.connect_output(node)
                    except Exception as e:
                        print(e)
                        return [] # Return an empty list to avoid printing None.

                # FIXME: Add support for io redirection; the parsing above supports it but it isn't
                # used yet.  For this we need some file nodes.
                if output_file:
                    output_node = output_file_node(output_file)
                    nodes.add(output_node)
                    node.connect_output(output_node)
                    nodes.add_output(output_node)
                
                last_node = node
                first_node = False

            # Suppress output from the pipeline if the last clause redirected output.
            if output_file:
                sink = sink_node()
                nodes.add(sink)
                node.connect_output(sink)
                node = sink
                
            nodes.add_output(node)
            return nodes

lingo.register('pipedsql', PipedSQLInterpreter)
