# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import textwrap
from boots.lib import _, n_

def issubseq(subseq, seq):
    """Determines if the elements the sequence subseq form a subsequence of the sequence seq."""
    for start in range(len(seq) - len(subseq) + 1):
        if seq[start:start + len(subseq)] == subseq:
            return True
    return False

def endswith(subseq, seq):
    return seq[len(seq) - len(subseq):] == subseq

def dedent(text):
    return "\n".join(line.strip() for line in text.strip().split("\n"))

def wrap(text, *args, **kwargs):
     return "\n".join(textwrap.fill(line,
                                    drop_whitespace=False,
                                    replace_whitespace=False,
                                    *args,
                                    **kwargs)
                      for line in text.split("\n"))   

def indent(text, initial=2*" ", subsequent=2*" "):
    return wrap(text,
                initial_indent=initial,
                subsequent_indent=subsequent)

class HelpTopic(object):
    """A topic in the help system which may contain other topics."""
    def __init__(self, name, title = None, short_description = None, description = None, see_also = None, aliases = None):
        """Creates a new a help topic.  The fields are as described above."""
        self._subtopics = dict()  
        self.title = title
        self.name = name
        self.short_description = short_description
        self._description = description or short_description
        if self._description:
            self._description = dedent(self._description)
        self.see_also = see_also
        self.aliases = aliases or []
        self.root = None

    @property
    def subtopics(self):
        return self._subtopics
    
    @property
    def description(self):
        if self._description is not None:
            return wrap(self._description)
        else:
            return None

    def format(self, include_index=False):
        """Returns a string that represents this topic in human readable form as
        described in detail above."""
        
        lines = []
        lines.append("=== {0} ===".format(self.title))
        lines.append(self.description)
        
        if self.aliases or self.see_also:
            lines.append("")
            
            if self.aliases:
                lines.append(_("Synonyms: {0}").format(", ".join(self.aliases)))
            
            if self.see_also:
                lines.append(_("See also: {0}").format(", ".join(self.see_also)))
        
        if include_index and self._subtopics:
            lines.append("--- {0} ---".format(_("Subtopics")))
            lines.append(self.format_index())
            
        return "\n".join(filter(lambda x: x is not None, lines))
 
    def format_index(self):
        """Returns a string that lists the subtopics of this topic in human
        readable form as described above."""
        lines = list()

        names = []
        for name, topic in self.subtopics.iteritems():
            names.append(", ".join([name] + topic.aliases))
        max_name_len = max(len(name) for name in names)
        
        for index, topic in enumerate(self.subtopics.itervalues()):
            subtext = "{name:{max_name_len}} : {description:30}".format(
                      name = names[index],
                      max_name_len = max_name_len,
                      description = topic.short_description)
            lines.append(subtext)
        lines.sort()
        return "\n".join(lines)

    # Queries:
    #
    # A query is used to refer to a topic in the help system.  A query is either
    # a string or a non-empty tuple of strings.  Using a string is equivalent to
    # using a tuple of length one containing that string.  As an example, the
    # tuple ('python', 'syntax', 'csv') refers to the csv subtopic in the syntax
    # topic which is itself a subtopic of the python topic.
 
    def _translate_query(self, query):
        """Converts a query into a tuple.  If query is not a valid query an
        exception is raised."""
        if isinstance(query, str):
            return (query,)
        elif isinstance(query, (tuple, list)):
            for item in query:
                assert isinstance(item, str)
            return query
        else:
            raise TypeError(_('query object {0} is not a string or tuple of strings').format(query))
 
    def _ensure_key(self, key):
        """Creates a new subtopic that key maps to if none exists."""
        if key not in self._subtopics:
            self._subtopics[key] = HelpTopic(key)
 
    def _walk(self, path=()):
        """Returns an iterable of all absolute paths."""
        for name, subtopic in self._subtopics.iteritems():
            for alias in [name]+subtopic.aliases:
                pos = path + (alias,)
                yield subtopic, pos
            for subsubtopic, subpath in subtopic._walk(pos):
                yield subsubtopic, subpath
 
    # The intention is for these methods to be implemented recursively
    # by calling these methods on topics contained in the current
    # topic (ie. this Help object).  Queries are represented as a
    # tuple denoting the topics.  For example, "python syntax csv"
    # would be represented as ('python', 'syntax', 'csv').
 
    def __contains__(self, query):
        """Return True if query refers to a help topic.  OTherwise,
        False is returned."""
        t = self._translate_query(query)
 
        if len(t) == 1:
            return t[0] in self._subtopics
        else:
            return t[0] in self._subtopics and t[1:] in self._subtopics[t[0]]
 
    def __getitem__(self, query):
        """Returns the help topic denoted by query; topics in query
        that do not exist should be created automatically.  query may
        refer to topic with subtopics."""
        t = self._translate_query(query)
        
        if not t:
            return self
        else:
            self._ensure_key(t[0])
     
            if len(t) == 1:
                return self._subtopics[t[0]]
            else:
                return self._subtopics[t[0]][t[1:]]
 
    def __setitem__(self, query, value):
        """Sets the help topic denoted by query to value.  topics in
        query that do not exist should be created automatically.
        query may refer to topic with subtopics."""
        t = self._translate_query(query)
        assert len(t) > 0
 
        if len(t) == 1:
            self._subtopics[t[0]] = value
            value.root = self.root if self.root is not None else self
        else:
            self._ensure_key(t[0])
            self._subtopics[t[0]][t[1:]] = value
 
    def __delitem__(self, query):
        """Deletes the reference to help topic denoted by query.
        query may refer to topic with subtopics."""
        t = self._translate_query(query)
        assert len(t) > 0
        
        if len(t) == 1:
            subtopic = self._subtopics[t[0]]
            subtopic.root = None
            del self._subtopics[t[0]]
        else:
            del self._subtopics[t[0]][t[1:]]
 
    def add(self, name, *args, **kwargs):
        kind = kwargs.pop("kind", None) or self.__class__
        self[name] = kind(name, *args, **kwargs)
 
    def search(self, query):
        """Returns an iterable of the absolute paths to topics that
        partially match query."""      
        if query:
            t = self._translate_query(query)
            for topic, path in self._walk():
                if endswith(t, path):
                    yield topic, path
        else:
            yield self, ()

class QueryIndexTopic(HelpTopic):
    def __init__(self, *args, **kwargs):
        self.query = kwargs.pop("query")
        HelpTopic.__init__(self, *args, **kwargs)
    
    @property
    def subtopics(self):
        subtopics = self._subtopics.copy()
        for topic, path in self.root.search(self.query):
            subtopics.update(topic._subtopics)
        return subtopics

class CommandHelpTopic(HelpTopic):
    def __init__(self, name, title = None, short_description = None, usage = None, description = None, see_also = None, aliases = None):
        HelpTopic.__init__(self, name, title, short_description, description, see_also, aliases)
        self.usage = usage
    
    @property
    def description(self):
        lines = []
        lines.append(_("Usage: ") + self.usage)
        if self._description:
            lines.append(_("Description:"))
            lines.append(indent(self._description))
        return "\n".join(lines)
                          