# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import warnings
import inspect
import types

from boots.api.errors import BootsError, BootsWarning
from boots.lib import _, n_

class MetaCommandError(BootsError):
    """Errors related to metacommands."""
    
class MetaCommandConflictWarning(BootsWarning):
    """Raised when conflicting metacommand definitions are encountered."""
    pass
    
class InvalidArgumentError(MetaCommandError):
    """Raised when a metacommand has an invalid argument."""

def argument_count_range(func):
    # Note: this can be fooled by things like *args
    argspec = inspect.getargspec(func)
    argcount = len(argspec.args)
    if type(func) is types.MethodType and func.im_self is not None:
        argcount -= 1
    mincount = argcount - len(argspec.defaults or [])
    
    if argspec.varargs is not None:
        return mincount, None
    else:
        return mincount, argcount

def parse_metacommand(commandstring):
    """This function wraps str.split() for metacommands

    This function takes a string as argument and calls str.split() on that string
    to return a list of substrings, using whitespace as separator. The metacommand
    object uses this list to call a metacommand function and pass parameters.
    
    """
    return commandstring.rstrip(";\n").split()    

class MetaCommandManager(object):
    def __init__(self):
        self.names = set()
        
    def add(self, metacommands):
        newnames = set(metacommands.commands.iterkeys())
        if not self.names.isdisjoint(newnames):
            # FIXME: Issue warning through the UI using ui.print
            # FIXME: Determine the offending command(s) (set.intersect?)
            # FIXME: Print offending commands
            warning.warn(MetaCommandConflictWarning(_("Conflicts between metacommands detected.")))
        self.names |= newnames

class MetaCommands(object):
    """Class used for handling metacommands.

    The MetaCommands class provides a mapping of command strings to function
    references. Its functions are the constructor, parse() (a function not bound
    to a MetaCommands object) and execute().
    
    The constructor accepts one parameter, a dictionary of string to function
    reference mappings. It stores this parameter in the member variable commands.
    
    """
    commands = {}

    def __init__(self, commanddict):
        self.commands = commanddict

    def get_metacommand(self, command):
        commandlist = parse_metacommand(command)
        if len(commandlist) != 0 and commandlist[0] in self.commands:
            return commandlist[0], self.commands[commandlist[0]], commandlist[1:]

    def execute(self, command):
        """Parse and execute a metacommand string.

        The execute() function takes a command string and parses out the name and 
        arguments. It performs a dictionary lookup on the name and calls the
        function it finds, or returns False if no function is found.
        
        """
        metacommand = self.get_metacommand(command)
        if metacommand:
            name, func, args = metacommand
            min_arg_count, max_arg_count = argument_count_range(func)
            if len(args) < min_arg_count or (max_arg_count is not None and len(args) > max_arg_count):
                if min_arg_count == max_arg_count:
                    raise InvalidArgumentError(n_("{0} takes {1} argument ({2} given)"
                                                  "{0} takes {1} arguments ({2} given)",
                                                  min_arg_count).format(name, min_arg_count, len(args)))
                else:
                    raise InvalidArgumentError(_("{0} takes {1}-{2} arguments ({3} given)")
                                               .format(name, min_arg_count, max_arg_count, len(args)))
            else:
                func(*args)
                # FIXME: Do we need to pass metacommand return values back?
                return True
        else:
            return False
