# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides a simple generic UI for boots. The UI is composed of
two pieces, each classes, the StreamDriver and StreamPresenter."""

import sys
import os
import StringIO

from boots.api.nodes.node import NodeGraph, SyncNode

class StreamDriver(object):
    """StreamDrivers receive stream inputs to boots, stdin for example."""
    def __init__(self, console, stream, lingo):
        """Initialize a StreamDriver giving it a link back to its parent
        console, the stream to receive input on, and the lingo being used."""
        self.console = console
        self.stream = stream
        self.lingo = lingo
        self.driver_graph = NodeGraph("streamdriver")
        self.driver_graph.add(SyncNode(self.get_input()), output=True)

    def get_input(self):
        """Gets user input from a stream."""
        try:
            lines = ""
            for line in self.stream:
                lines += line
                command = self.console.input_complete(lines, self.lingo)
                if command:
                    yield command, self.lingo
                    lines = ""
                                
        except EOFError:
            return
        
    @property
    def is_interactive(self):
        if hasattr(os, "isatty") and hasattr(self.stream, "fileno"):
            return os.isatty(self.stream.fileno())
        else:
            return False

class StreamPresenter(object):
    """StreamPresenters write out boots results to streams, stdout for
    example."""
    def __init__(self, stream):
        """Initialize a StreamPresenter giving it the stream to write to."""
        self.stream = stream
        self.presenter_graph = NodeGraph("streampresenter")
        self.presenter_graph.add(SyncNode(self.present), input=True, output=True)
        
    def present(self, row):
        """Prints row information in a "raw" format"""
        self.stream.write(str(row))
        self.stream.write("\n")
        
class StdinDriver(StreamDriver):
    def __init__(self, console, lingo):
        super(StdinDriver, self).__init__(sys.stdin, console, lingo)

class StdoutPresenter(StreamPresenter):
    def __init__(self,):
        super(StdoutPresenter, self).__init__(sys.stdout)
        
class ScriptDriver(StreamDriver):
    def __init__(self, console, path, lingo):
        self.path = path
        self.script_file = open(path)
        super(ScriptDriver, self).__init__(console, self.script_file, lingo)
        
class StringDriver(StreamDriver):
    def __init__(self, console, input_string, lingo):
        self.input_string = input_string
        string_stream = StringIO.StringIO(self.input_string)
        super(StringDriver, self).__init__(console, string_stream, lingo)
