# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module provides a 'plain' UI for boots. The UI is plain because it
uses stdin and stdout to do its job."""

import sys
import os
import time
import readline
import subprocess
import threading
from itertools import izip

from boots.api.api import Rows, ResultInfo
from boots.api.nodes.node import Status, NodeGraph, SyncNode, QueueNode, filter_none
from boots.lib.ui.components.help import HelpTopic, CommandHelpTopic
from boots.lib.ui.components.metacommands import MetaCommandError, MetaCommands, parse_metacommand
from boots.lib.ui.generic import StdinDriver, StdoutPresenter
from boots.lib import _, n_

class PlainUI(StdinDriver, StdoutPresenter):
    """Class that provides a 'plain' UI for boots. Input is taken on stdin,
    primary and secondary prompts are provided depending on context, and
    results are printed to stdout. These results are printed as tables when
    they are Server packets and as string interpretations otherwise."""
    
    help = HelpTopic("plain", _("Plain UI"), _("Plain UI documentation."),
        description = _("A simple, minimal user interface for Boots."))
    
    help.add("commands", None, _("List of Plain UI commands."))
    help["commands"].add("\use", _("\use"), _("Switch to a different lingo."),
        usage       = "\use lingo",
        description = _("Switch the selected lingo. Future commands entered will be executed in the specified lingo."),
        kind = CommandHelpTopic)

#    help["commands"].add("historylength", _("historylength"), _("Specify maximum history file length."))    
#    help["commands"].add("historyfile", _("historyfile"), _("Uses readline to provide console history persistently."))
    def __init__(self, console):
        """Initialize a UI giving it its parent console, primary prompt,
        secondary prompt, maximum number of history lines, history file, and
        lingo to be used."""
        self.metacommands = MetaCommands({
            '\use': self.switchlingo})
        self.console = console
        console.hooks["unload"].register(self.unload)
        
        self.prompt1 = console.config["prompt1"]
        self.prompt2 = console.config["prompt2"]
        self.hist_file = console.config["history_file"]
        self.lingo = console.config["lingo"]
        self.pager_command = console.config["pager_command"]
        self.last_desc = None
        self.buffer = []
        
        self.driver_graph = NodeGraph("plaindriver")
        self.driver_graph.add(SyncNode(self.get_input(), name="get_input"), output=True)
        
        self.presenter_graph = NodeGraph("plainpresenter")
        self.presenter_graph.add(SyncNode(self.present, name="present", filter=filter_none), input=True, output=True)
                
        readline.set_history_length(console.config["history_length"])
        self.read_history()

    def read_history(self):
        """Read existing history from the history file."""
        try:
            readline.read_history_file(self.hist_file)
        except IOError:
            pass

    def save_history(self):
        """Save history from this session back to the history file."""
        # FIXME: commands with newlines are not properly being preserved
        readline.write_history_file(self.hist_file)
        
    def unload(self):
        """'unload' a PlainUI saving its history."""
        self.save_history()
        sys.stdout.flush()

    def get_input(self):
        """Gets user input from stdin via a prompt.
        
        On incomplete inputs print the secondary prompt instead of the
        primary print."""
        while True:
            try:
                line = raw_input(self.prompt1)
                if not self.metacommands.execute(line):
                    lines = line + "\n"
                    command = self.console.input_complete(line, self.lingo)
                    while command is None:
                        line = raw_input(self.prompt2)
                        lines += line + "\n"
                        command = self.console.input_complete(lines, self.lingo)

                    yield command, self.lingo
                
            except KeyboardInterrupt:
                # FIXME: This is a hack to emulate bash's behavior
                # ^C/SIGINT should kill the current Console job, but that will require
                # some thinking. 
                print "^C"
                continue
            except EOFError:
                return
            except MetaCommandError as e:
                yield e
            
    def set_prompt(self, prompt1 = None, prompt2 = None):
        """Change the prompts during execution of a PlainUI."""
        if prompt1 is None and prompt2 is None:
            return
        elif prompt2 is None:
            self.prompt1 = prompt1
        elif prompt1 is None:
            self.prompt2 = prompt2
        else:
            self.prompt1 = prompt1
            self.prompt2 = prompt2

    def print_with_pager(self, text):
        try:
            pager = subprocess.Popen(self.pager_command.split(),
                                     shell=False,
                                     stdin=subprocess.PIPE)
        except:
            sys.stdout.write(_("Unable to run pager command \"{0}\". Pager disabled.\n")
                             .format(self.pager_command))
            self.pager_command = None
            return False
            
        try:
            pager.communicate(text)
        except IOError:
            # IOError is raised sometimes with large outputs
            pass
        
        return True

    def present(self, result):
        """Print the result provided as an argument.

        If the result is a packet buffer the results until the last packet is
        received. At that point print all of the buffered result rows in table
        format. Non packet results are presented in their string representation
        as received."""
        def padded(fields, widths):
            """Utility function used to convert rows from tuples to table rows.
            Results are returned as strings."""
            padded_fields = []
            for index, field in enumerate(fields):
                padded_fields.append(field.ljust(widths[index]))
            return u"| {0} |".format(u" | ".join(padded_fields))
        
        def show_NULL(value):
            """There is a 'bug' in the dbapi that does not convert NULL objects
            to the string 'NULL'. This utility function performs that
            conversion."""
            return value if value is not None else "NULL"

        def _gen_table(info):
            if self.buffer:
                max_widths = map(max, [(len(column[0]), column[2]) for column in info["description"]])
                dashes = map(lambda x: "-"*(x+2), max_widths)
                sep_line = "+" + "+".join(dashes) + "+"
                names = (column[0] for column in info["description"])
                yield sep_line
                yield padded(names, max_widths)
                yield sep_line
                for row in self.buffer:
                    yield padded(map(show_NULL, row), max_widths)
    
                yield sep_line
            
            if self.console.driver.is_interactive:
                elapsed = info["end_time"] - info["begin_time"]
                noun = "row" if info["row_count"] == 1 else "rows"
                if info["description"] is not None:
                    info_line = n_("{count} row in set ({elapsed:.2f} seconds)", 
                                   "{count} rows in set ({elapsed:.2f} seconds)",
                                   info["row_count"])
                else:
                    info_line = n_("{count} row affected ({elapsed:.2f} seconds)",
                                   "{count} rows affected ({elapsed:.2f} seconds)",
                                   info["row_count"])
                    
                yield info_line.format(count=info["row_count"],
                                       noun=noun,
                                       elapsed=elapsed)

        if type(result) is Rows:
            self.buffer.extend(result)
        elif isinstance(result, ResultInfo):
            printed = False
            output = _gen_table(result.value)
            if self.pager_command and self.console.driver.is_interactive:
                printed = self.print_with_pager("\n".join(output).encode("utf-8"))
            
            # Fallback if no pager set, or paging fails.
            if not printed:
                for line in output:
                    sys.stdout.write(line.encode("utf-8"))
                    sys.stdout.write("\n")
                
            # Reset values for next result set.
            self.buffer = []
        elif isinstance(result, Status):
            sys.stdout.write(_("Status"))
            sys.stdout.write(" :: ")
            sys.stdout.write(str(result))
            sys.stdout.write("\n")
        elif isinstance(result, Warning):
            sys.stderr.write(_("WARNING"))
            sys.stderr.write(" :: ")
            sys.stderr.write(str(result))
            sys.stderr.write("\n")
        elif isinstance(result, Exception):
            sys.stderr.write(_("ERROR"))
            sys.stderr.write(" :: ")
            sys.stderr.write(str(result))
            sys.stderr.write("\n")
        elif result is not None:
            sys.stdout.write(str(result))
            sys.stdout.write("\n")
                
        if isinstance(result, Exception):
            return result
        else:
            return True

    @property
    def is_interactive(self):
        if hasattr(os, "isatty"):
            return os.isatty(sys.stdin.fileno())
        else:
            return True

    def switchlingo(self, lingo):
        """Switch the lingo used by the PlainUI during execution to the
        lingo provided as an argument."""
        if lingo in self.console.lingos:
            self.lingo = lingo
        else:
            self.present(_("The specified lingo \"{0}\" does not exist.").format(lingo))
