# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

"""This module contains unit tests for the boots' api module."""
import unittest
from itertools import ifilter

from boots.api import errors, api
from boots.api.nodes.node import filter_default
import boots_unit_test

class TestServerConnections(boots_unit_test.BootsQueryTester):
    """Class that provides unit tests that test the api.DrizzleServer class' ability
    to connect to servers as well as provide proper errors when connections
    should fail."""
    #FIXME should use the configuration provided.
    def setUp(self):
        #We want to derive from the BootsQueryTester but not use the default
        #setUp method due to the cost incurred by it.
        pass

    def test_valid_connect_fqdn(self):
        server = api.DrizzleServer("capstonedd.cs.pdx.edu", 9306, {})
        server.connect()
        self.assert_(server.is_connected)
        server.disconnect()

    def test_valid_connect_partial(self):
        server = api.DrizzleServer("capstonedd", 9306, {})
        server.connect()
        self.assert_(server.is_connected)
        server.disconnect()

    def test_valid_connect_ip(self):
        server = api.DrizzleServer("131.252.214.178", 9306, {})
        server.connect()
        self.assert_(server.is_connected)
        server.disconnect()

    def test_valid_connect_fqdn_with_db(self):
        boots_unit_test.BootsQueryTester.setUp(self)
        server = api.DrizzleServer("capstonedd.cs.pdx.edu", 9306, {"database": "fec"})
        server.connect()
        self.assert_(server.is_connected)
        server.disconnect()

    def test_invalid_connect(self):
        server = api.DrizzleServer("kororaa.cs.pdx.edu", 9306, {})
        self.assertRaises(errors.ConnectionError, server.connect)
        self.assert_(not server.is_connected)

    def test_valid_disconnect(self):
        server = api.DrizzleServer("capstonedd.cs.pdx.edu", 9306, {})
        server.connect()
        server.disconnect()
        self.assert_(not server.is_connected)

    def test_invalid_disconnect(self):
        server = api.DrizzleServer("kororaa.cs.pdx.edu", 9306, {})
        self.assertRaises(errors.ConnectionError, server.connect)
        self.assertRaises(errors.ConnectionError, server.disconnect)
        self.assert_(not server.is_connected)

class TestServerExecute(boots_unit_test.BootsQueryTester):
    """Class that provides unit tests that tests api.DrizzleServer class' ability
    to execute SQL statements on a server."""
    #FIXME needs to test more than just SELECT statements.
    def setUp(self):
        boots_unit_test.BootsQueryTester.setUp(self)
        self.server = api.DrizzleServer(self.config["host"], self.config["port"], {"database": self.config["database"]})
        self.server.connect()

    def tearDown(self):
        self.server.disconnect()

    def _concat_packet_results(self, iter):
        """Results come from the Server object in 'packets'. These packets need
        their rows extracted so that they can be compared properly. This method
        takes an iterator of packets and does this extraction for you. Results
        are returned in a single list."""
        result = []
        for data in ifilter(filter_default, iter):
            result.extend(data)
        return result

    def test_query_length_0(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM cand WHERE name = 'foobar';"))
        self.assertEqual(len(result), 0)

    def test_query_length_1(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM comm WHERE name = 'FUNK FOR CONGRESS';"))
        self.assertEqual(len(result), 1)

    def test_query_length_914(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM cand, comm WHERE cand.id = comm.cand_id;"))
        self.assertEqual(len(result), 914)

    def test_query_length_1710(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM comm;"))
        self.assertEqual(len(result), 1710)

    def test_query_length_2501730(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM cand,comm;"))
        self.assertEqual(len(result), 2501730)

    def test_query_values_cand(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM cand WHERE name = 'SPACE, ZACHARY T';"))
        self.assertEqual(result[0], 
                         ("H6OH18157",
                          "SPACE, ZACHARY T",
                          "H",
                          "OH",
                          "DEM",
                          "4 PARKVIEW DRIVE", 
                          "DOVER", 
                          "OH",
                          "44622"))

    def test_query_values_comm(self):
        result = self._concat_packet_results(self.server.execute("SELECT * FROM comm WHERE name = 'COMMITTEE TO RE-ELECT TOM FOLEY';"))
        self.assertEqual(result[0],
                         ("COMMITTEE TO RE-ELECT TOM FOLEY",
                          "1253 C STREET SE BOX 200",
                          "WASHINGTON",
                          "DC",
                          "20003",
                          "C00002592",
                          "2010",
                          "H6WA05023"))

    def test_query_identity_same_q(self):
        result1 = self._concat_packet_results(self.server.execute("SELECT * FROM cand;"))
        result2 = self._concat_packet_results(self.server.execute("SELECT * FROM cand;"))
        self.assertEqual(result1, result2)

    def test_query_identity_equiv_q(self):
        result1 = self._concat_packet_results(self.server.execute("SELECT * FROM comm;"))
        result2 = self._concat_packet_results(self.server.execute("SELECT name, street, city, state, zipcode, id, year, cand_id FROM comm;"))
        self.assertEqual(result1, result2)

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestServerConnections)
    unittest.TextTestRunner(verbosity=2).run(suite)
    suite = unittest.TestLoader().loadTestsFromTestCase(TestServerExecute)
    unittest.TextTestRunner(verbosity=2).run(suite)
