# Boots Client Project
# www.launchpad.net/boots
#
# ##### BEGIN LICENSE BLOCK #####
#
# Copyright (C) 2009-2010 Clark Boylan, Ken Brotherton, Max Goodman,
#                         Victoria Lewis, David Rosenbaum, and Andreas Turriff
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END LICENSE BLOCK #####

import unittest
import boots_unit_test
from boots.api.nodes.node import NodeGraph, SyncNode
from boots.app import client_config
from boots.lib import console
from boots.lib.lingos import lingo, bash_external, lisp, piped_sql, python, sql
import os
import tempfile

def node_graph_values(graph):
    values = []
    output_graph = NodeGraph()

    def append(obj):
        values.append(obj)
        
    output_node = SyncNode(append)
    output_graph.add_input(output_node)
    output_graph.add_output(output_node)
    graph.link_output(output_graph)
    graph.start()
    graph.run()
    output_graph.start()
    output_graph.run()
    return values

class LingoBaseTester(boots_unit_test.BootsBaseTester):
    def setUp(self):
        self.config = client_config.ClientConfig()
        self.console = console.Console(self.config)
        self.console.connect('capstonedd.cs.pdx.edu', 9306, 'fec')
        
    def tearDown(self):
        pass

class TestLingo(LingoBaseTester):
    def setUp(self):
        pass

    def tearDown(self):
        pass
    
    def test_example(self):
        self.assert_(False)

class TestBashExternal(LingoBaseTester):
    def setUp(self):
        super(TestBashExternal, self).setUp()
        self.bash = bash_external.ExternalBashInterpreter(self.console)
        
    def test_plus(self):
        result = self.bash.evaluate('echo $(( 1 + 1 ))')[0]
        self.assertEqual(result, '2')

class TestSQL(LingoBaseTester):
    def setUp(self):
        super(TestSQL, self).setUp()
        self.sql = sql.SQLInterpreter(self.console)

    def test_plus(self):
        result = node_graph_values(self.sql.execute('select 1 + 1;'))[0][0][0]
        self.assertEqual(result, '2')

class TestLisp(LingoBaseTester):
    def setUp(self):
        super(TestLisp, self).setUp()
        self.lisp = lisp.lisp.LispInterpreter(self.console)

    def test_plus(self):
        result = self.lisp.execute('(+ 1 2 3 4)').python_integer()
        self.assertEqual(result, 10)

    def test_recursion(self):
        self.lisp.execute('(setq len (lambda (l) (if l (+ 1 (len (cdr l))) 0)))')
        result = self.lisp.execute('(len \'(1 2 3 4))').python_integer()
        self.assertEqual(result, 4)
        result = self.lisp.execute('(len nil)').python_integer()
        self.assertEqual(result, 0)

class TestPipedSQLInterpreter(TestSQL):
    def setUp(self):
        super(TestPipedSQLInterpreter, self).setUp()
        self.sql = piped_sql.PipedSQLInterpreter(self.console)

    def test_redirect(self):
        temp = tempfile.mkstemp(prefix = 'boots-test-file')[1]
        piped_sql_query = 'select * from cand; > {0}'.format(temp)
        graph = self.sql.execute(piped_sql_query)
        graph.start()
        graph.run()
        self.assertEqual(len(open(temp).readlines()), 1458)

class TestPythonInterpreter(LingoBaseTester):
    def setUp(self):
        super(TestPythonInterpreter, self).setUp()
        self.python = python.PythonInterpreter(self.console)

    def test_plus(self):
        result = self.python.execute('1 + 2 + 3')
        self.assertEqual(result, 6)
    
    def test_len(self):
        result = self.python.execute('len([1,2,3,4])')
        self.assertEqual(result, 4)
